/*
Very simple GLUT program using OSL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2006/08/31 (Public Domain)
*/
#include "ogl/ext.h" /* OpenGL Extensions, for programmable shaders */
#include "ogl/main.h" /* My main program */
#include "osl/viewpoint.h" /* 3D viewpoint utility routines */
#include "osl/viewpoint.cpp" 
#include "osl/random.h" /* random number routines */
#include "osl/random.cpp" 
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <vector>

/************ Main Drawing Function **************/

/**
  Program to run on each vertex or pixel fragment.
*/
class oglShaderObject {
	int target;// Program target, like GL_FRAGMENT_SHADER_ARB
	GLhandleARB progName; // OpenGL handle to program
public:
	oglShaderObject(int target_) :target(target_) 
		{progName=glCreateShaderObjectARB(target);}
	
	/// Get our handle (e.g., for glAttachObjectARB)
	GLhandleARB get(void) const {return progName;}
	
	// Replace the current program by the one in this file.
	//  Prints out an error message if the program can't be parsed.
	void read(const char *fName);
	
	// Replace the current program by this one.
	//  Prints out an error message if the program can't be parsed.
	void set(const char *programText);
	
	~oglShaderObject() {glDeleteObjectARB(progName);}
};

void oglShaderObject::read(const char *fName) {
	
	// Read program from file
	int len=65536;
	char *s=new char[len];
	FILE *f=fopen(fName,"r");
	len=fread(s,1,len,f); // reset len to file length
	s[len]=0; // null terminate string
	fclose(f);
	
	// Upload program to graphics card
	set(s);
	
	delete[] s;
}

void checkOp(GLhandleARB obj,int what,const char *where) {
	GLint compiled;
	glGetObjectParameterivARB(obj,what,&compiled);
	if (!compiled) {
		printf("Compile error on program: %s\n",where);
		enum {logSize=1000};
		char log[logSize]; int len=0;
		glGetInfoLogARB(obj, logSize,&len,log);
		printf("Log: \n%s\n",log);
		exit(1);
	}
}
	
void oglShaderObject::set(const char *programText) {
	glShaderSourceARB(progName,1, &programText,NULL);
	glCompileShaderARB(progName);
	checkOp(progName,GL_OBJECT_COMPILE_STATUS_ARB,programText);
	oglCheck(programText);
}
void uploadShaders(oglShaderObject &vert,oglShaderObject &frag,
	GLhandleARB &prog,const oglViewpoint &vp) 
{
	double start_upload=oglTimer();
	
	vert.read("vertex.txt");
	frag.read("fragment.txt");
	
	glAttachObjectARB(prog,vert.get());
	glAttachObjectARB(prog,frag.get());

	glLinkProgramARB(prog);
	checkOp(prog,GL_OBJECT_LINK_STATUS_ARB,"link");
	oglCheck("link");
	
	if (0) { /* Linker log can show what runs on hardware */
		char log[65536];int loglen=0;
		glGetInfoLogARB(prog,sizeof(log)-1,&loglen,log);
		log[loglen]=0;
		printf("Log: %s\n",log);
	}
	
	glUseProgramObjectARB(prog);
	oglCheck("use");
	
	/* Set up texture "sampler" as corresponding texture unit */
	glUniform1iARB(glGetUniformLocationARB(prog,"tex0"),0);
	glUniform1iARB(glGetUniformLocationARB(prog,"tex1"),1);
	glUniform1iARB(glGetUniformLocationARB(prog,"tex2"),2);
	
	double elapsed_upload=oglTimer()-start_upload;
	if (oglKey('t',"Print program compile time"))
		printf("Upload: %.2f ms\n",1.0e3*elapsed_upload);
}


class myVertex {
public:
	osl::Vector3f loc; /* Location of vertex */
	osl::Vector3f normal; /* Normal of vertex */
	//osl::Vector3f color; /* Color of vertex */
	myVertex(const osl::Vector3f &l,const osl::Vector3f &n) :loc(l),normal(n) {}
};
class myTriangle {
public:
	int v[3];
	myTriangle(int i,int j,int k) {v[0]=i;v[1]=j;v[2]=k;}
};

class fragBench : public oglDrawable {
public:
	std::vector<myVertex> vertices;
	GLuint vertex_buffer;
	std::vector<myTriangle> triangles;
	GLuint triangle_buffer;
	
	fragBench();
	virtual void draw(const oglViewpoint &vp);
};

/******** Program Initialization *********/
int main(int argc, char** argv)
{
	oglOptions i;
	i.width=i.height=700;
	i.name="GL_ARB_shader_language_100 Test";
	i.clearColor=oglVector3d(0.2,0.3,0.4);
	oglSetup(i);
	oglSetupExt();
	if (!GLEW_ARB_fragment_shader) {
		printf("You need programmable shading to run this program!\n");
		exit(1);
	}

	oglEventloop(new oglTrackballController(),new fragBench());
	
	return 0;
}

/* Silly little height function */
double getHeight(double x,double y) {
	return
		0.10*cos(125*(x*x+y*y))+
		0.05*cos(137*((x+0.1)*(x+0.1)+y*y));
}
osl::Vector3f makeVertex(double x,double y) {
	return osl::Vector3f(x,y,getHeight(x,y));
}
osl::Vector3f makeNormal(double x,double y) {
	osl::Vector3f c=makeVertex(x,y);
	osl::Vector3f a=makeVertex(x+0.0001,y);
	osl::Vector3f b=makeVertex(x,y+0.0001);
	return (a-c).cross(b-c).dir();
}


/** Set up a 3D model, stored in a vertex buffer */
fragBench::fragBench()
{

  if (1) { /* a simple 2D plane of points */
	int n=100; /* 10000 points: 100 x 100 array */
	if (1) n=1000; /* 1m points: 1000 x 1000 array! */
	double sz=1.0/n;
	for (int iy=0;iy<n;iy++) {
		double y=iy*sz;
		for (int ix=0;ix<n;ix++) {
			double x=ix*sz;
			vertices.push_back(
				myVertex(makeVertex(x,y),makeNormal(x,y))
			);
		}
		
	}
  }
  else { /* Sierpinski gasket */
  	osl::Vector3d v(0,0,0);
	osl::Random48 r;
  	for (int point=0;point<1000000;point++) {
		v=v*0.5; /* scale down */
		switch (r.nextInt(4)) {
		case 0: break;
		case 1: v+=osl::Vector3d(0,0.5,0); break;
		case 2: v+=osl::Vector3d(0.5,0.25,0); break;
		case 3: v+=osl::Vector3d(0.25,0.25,0.5); break;
		}
		vertices.push_back(myVertex(v,osl::Vector3d(0,1,0)));
	}
  }
	
	vertex_buffer=0;
	glGenBuffersARB(1, &vertex_buffer);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,
		 sizeof(myVertex)*vertices.size(), 
		 &vertices[0], 1?GL_STATIC_DRAW_ARB:GL_STREAM_DRAW_ARB);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
}

char *thingy=0;
#define my_offsetof(my_struct,my_field) \
	(int)(((char *)&(((my_struct *)thingy)->my_field))-thingy)

void fragBench::draw(const oglViewpoint &vp)
{
	oglShaderObject vert(GL_VERTEX_SHADER_ARB);   
	oglShaderObject frag(GL_FRAGMENT_SHADER_ARB); 
	
	GLhandleARB prog=glCreateProgramObjectARB();
	if (!oglKey('f',"Disable vertex and fragment shaders")) uploadShaders(vert,frag,prog,vp);
	
	/* Alpha blending requires a valid alpha channel, which can be tricky when using a pixel shader */
	glDisable(GL_BLEND);
	
	/* Bind in array of points */
	
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffer);
	glEnableClientState(GL_VERTEX_ARRAY);	
	glVertexPointer(3,GL_FLOAT,
		sizeof(myVertex),(void *)my_offsetof(myVertex,loc));
	glEnableClientState(GL_NORMAL_ARRAY);	
	glNormalPointer(GL_FLOAT,
		sizeof(myVertex),(void *)my_offsetof(myVertex,normal));
/*
	glEnableClientState(GL_COLOR_ARRAY);	
	glColorPointer(3,GL_FLOAT,
		sizeof(myVertex),(void *)my_offsetof(myVertex,color));
*/
	
	//glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size());
	//glDrawElements(GL_TRIANGLES,3*triangles.size(),
	//	GL_UNSIGNED_INT,0);
	glDrawArrays(GL_POINTS,0,vertices.size());
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	
	oglRepaint(10);
	
	glUseProgramObjectARB(0);
	glDeleteObjectARB(prog);
}
